@@ -1,9 +1,10 @@
 #
 # python-netsnmpagent module
-#
-# Copyright (c) 2012 Pieter Hollants <pieter@hollants.com>
+# Copyright (c) 2013 Pieter Hollants <pieter@hollants.com>
 # Licensed under the GNU Public License (GPL) version 3
 #
+# Main module
+#
 
 """Allows to write net-snmp subagents in Python.
 
@@ -19,6 +20,7 @@ for SNMP subagents in an easy manner. It is still under heavy
 development and some features are yet missing."""
 
 import sys, os, socket, struct
+from collections import defaultdict
 from netsnmpapi import *
 
 # Maximum string size supported by python-netsnmpagent
@@ -33,26 +35,26 @@ class netsnmpAgent(object):
 		"args" is a dictionary that can contain the following
 		optional parameters:
 		
-		- AgentName    : The agent's name used for registration with net-snmp.
-		- MasterSocket : The Unix domain socket of the running snmpd instance to
-		                 connect to. Change this if you want to use a custom
-		                 snmpd instance, eg. in example.sh or for automatic
-		                 testing.
-		- PersistentDir: The directory to use to store persistance information.
-		                 Change this if you want to use a custom snmpd instance,
-		                 eg. in example.sh or for automatic testing.
-		- MIBFiles     : A list of filenames of MIBs to be loaded. Required if
-		                 the OIDs, for which variables will be registered, do
-		                 not belong to standard MIBs and the custom MIBs are not
-		                 located in net-snmp's default MIB path
-		                 (/usr/share/snmp/mibs). """
+		- AgentName     : The agent's name used for registration with net-snmp.
+		- MasterSocket  : The Unix domain socket of the running snmpd instance
+		                  to connect to. Change this if you want to use a custom
+		                  snmpd instance, eg. in example.sh or for automatic
+		                  testing.
+		- PersistenceDir: The directory to use to store persistence information.
+		                  Change this if you want to use a custom snmpd
+		                  instance, eg. for automatic testing.
+		- MIBFiles      : A list of filenames of MIBs to be loaded. Required if
+		                  the OIDs, for which variables will be registered, do
+		                  not belong to standard MIBs and the custom MIBs are not
+		                  located in net-snmp's default MIB path
+		                  (/usr/share/snmp/mibs). """
 
 		# Default settings
 		defaults = {
-			"AgentName"    : os.path.splitext(os.path.basename(sys.argv[0]))[0],
-			"MasterSocket" : None,
-			"PersistentDir": None,
-			"MIBFiles"     : None
+			"AgentName"     : os.path.splitext(os.path.basename(sys.argv[0]))[0],
+			"MasterSocket"  : None,
+			"PersistenceDir": None,
+			"MIBFiles"      : None
 		}
 		for key in defaults:
 			setattr(self, key, args.get(key, defaults[key]))
@@ -84,11 +86,11 @@ class netsnmpAgent(object):
 				)
 
 		# Use an alternative persistence directory?
-		if self.PersistentDir:
+		if self.PersistenceDir:
 			if libnsa.netsnmp_ds_set_string(
 				NETSNMP_DS_LIBRARY_ID,
 				NETSNMP_DS_LIB_PERSISTENT_DIR,
-				ctypes.c_char_p(self.PersistentDir)
+				ctypes.c_char_p(self.PersistenceDir)
 			) != SNMPERR_SUCCESS:
 				raise netsnmpAgentException(
 					"netsnmp_ds_set_string() failed for NETSNMP_DS_LIB_PERSISTENT_DIR!"
@@ -112,7 +114,7 @@ class netsnmpAgent(object):
 					                            "failed!".format(mib))
 
 		# Initialize our SNMP object registry
-		self._objs    = {}
+		self._objs    = defaultdict(dict)
 		self._started = False
 
 	def _prepareRegistration(self, oidstr, writable = True):
@@ -176,12 +178,14 @@ class netsnmpAgent(object):
 		                          eg. 0 or "".
 		    - "asntype"         : A constant defining the SNMP variable type
 		                          from an ASN.1 perspective, eg. ASN_INTEGER.
+		    - "context"         : A string defining the context name for the
+		                          SNMP variable
 		
 		    The class instance returned will have no association with net-snmp
 		    yet. Use the Register() method to associate it with an OID. """
 
 		# This is the replacement function, the "decoration"
-		def create_vartype_class(self, oidstr = None, initval = None, writable = True):
+		def create_vartype_class(self, initval = None, oidstr = None, writable = True, context = ""):
 			agent = self
 
 			# Call the original property_func to retrieve this variable type's
@@ -223,19 +227,19 @@ class netsnmpAgent(object):
 					if oidstr:
 						# Prepare the netsnmp_handler_registration structure.
 						handler_reginfo = agent._prepareRegistration(oidstr, writable)
+						handler_reginfo.contents.contextName = context
 
 						# Create the netsnmp_watcher_info structure.
-						watcher = libnsa.netsnmp_create_watcher_info6(
+						watcher = libnsX.netsnmp_create_watcher_info(
 							self.cref(),
 							self._data_size,
 							self._asntype,
-							self._flags,
-							self._max_size,
-							None
+							self._flags
 						)
+						watcher.max_size = self._max_size
 
 						# Register handler and watcher with net-snmp.
-						result = libnsa.netsnmp_register_watched_instance(
+						result = libnsX.netsnmp_register_watched_scalar(
 							handler_reginfo,
 							watcher
 						)
@@ -244,16 +248,23 @@ class netsnmpAgent(object):
 
 						# Finally, we keep track of all registered SNMP objects for the
 						# getRegistered() method.
-						agent._objs[oidstr] = self
+						agent._objs[context][oidstr] = self
 
 				def value(self):
-					return self._cvar.value
+					val = self._cvar.value
+					if val <= sys.maxint:
+						val = int(val)
+					return val
 
 				def cref(self):
 					return ctypes.byref(self._cvar) if self._flags == WATCHER_FIXED_SIZE \
 					                                else self._cvar
 
 				def update(self, val):
+					if self._asntype == ASN_COUNTER and val >> 32:
+						val = val & 0xFFFFFFFF
+					if self._asntype == ASN_COUNTER64 and val >> 64:
+						val = val & 0xFFFFFFFFFFFFFFFF
 					self._cvar.value = val
 					if props["flags"] == WATCHER_SIZE_STRLEN:
 						if len(val) > self._max_size:
@@ -271,7 +282,7 @@ class netsnmpAgent(object):
 		return create_vartype_class
 
 	@VarTypeClass
-	def Integer32(self, oidstr = None, initval = None, writable = True):
+	def Integer32(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : ctypes.c_long,
 			"flags"         : WATCHER_FIXED_SIZE,
@@ -280,7 +291,7 @@ class netsnmpAgent(object):
 		}
 
 	@VarTypeClass
-	def Unsigned32(self, oidstr = None, initval = None, writable = True):
+	def Unsigned32(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : ctypes.c_ulong,
 			"flags"         : WATCHER_FIXED_SIZE,
@@ -289,7 +300,7 @@ class netsnmpAgent(object):
 		}
 
 	@VarTypeClass
-	def Counter32(self, oidstr = None, initval = None, writable = True):
+	def Counter32(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : ctypes.c_ulong,
 			"flags"         : WATCHER_FIXED_SIZE,
@@ -298,7 +309,7 @@ class netsnmpAgent(object):
 		}
 
 	@VarTypeClass
-	def Counter64(self, oidstr = None, initval = None, writable = True):
+	def Counter64(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : counter64,
 			"flags"         : WATCHER_FIXED_SIZE,
@@ -307,7 +318,7 @@ class netsnmpAgent(object):
 		}
 
 	@VarTypeClass
-	def TimeTicks(self, oidstr = None, initval = None, writable = True):
+	def TimeTicks(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : ctypes.c_ulong,
 			"flags"         : WATCHER_FIXED_SIZE,
@@ -318,7 +329,7 @@ class netsnmpAgent(object):
 	# Note we can't use ctypes.c_char_p here since that creates an immutable
 	# type and net-snmp _can_ modify the buffer (unless writable is False).
 	@VarTypeClass
-	def OctetString(self, oidstr = None, initval = None, writable = True):
+	def OctetString(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : ctypes.create_string_buffer,
 			"flags"         : WATCHER_SIZE_STRLEN,
@@ -330,7 +341,7 @@ class netsnmpAgent(object):
 	# Whereas an OctetString can contain UTF-8 encoded characters, a
 	# DisplayString is restricted to ASCII characters only.
 	@VarTypeClass
-	def DisplayString(self, oidstr = None, initval = None, writable = True):
+	def DisplayString(self, initval = None, oidstr = None, writable = True, context = ""):
 		return {
 			"ctype"         : ctypes.create_string_buffer,
 			"flags"         : WATCHER_SIZE_STRLEN,
@@ -341,7 +352,7 @@ class netsnmpAgent(object):
 
 	# IP addresses are stored as unsigned integers, but the Python interface
 	# should use strings. So we need a special class.
-	def IpAddress(self, oidstr = None, initval = "0.0.0.0", writable = True):
+	def IpAddress(self, initval = "0.0.0.0", oidstr = None, writable = True, context = ""):
 		agent = self
 
 		class IpAddress(object):
@@ -352,19 +363,19 @@ class netsnmpAgent(object):
 				if oidstr:
 					# Prepare the netsnmp_handler_registration structure.
 					handler_reginfo = agent._prepareRegistration(oidstr, writable)
+					handler_reginfo.contents.contextName = context
 
 					# Create the netsnmp_watcher_info structure.
-					watcher = libnsa.netsnmp_create_watcher_info6(
+					watcher = libnsX.netsnmp_create_watcher_info(
 						self.cref(),
 						ctypes.sizeof(self._cvar),
 						ASN_IPADDRESS,
-						WATCHER_FIXED_SIZE,
-						ctypes.sizeof(self._cvar),
-						None
+						WATCHER_FIXED_SIZE
 					)
+					watcher._maxsize = ctypes.sizeof(self._cvar)
 
 					# Register handler and watcher with net-snmp.
-					result = libnsa.netsnmp_register_watched_instance(
+					result = libnsX.netsnmp_register_watched_instance(
 						handler_reginfo,
 						watcher
 					)
@@ -373,7 +384,7 @@ class netsnmpAgent(object):
 
 					# Finally, we keep track of all registered SNMP objects for the
 					# getRegistered() method.
-					agent._objs[oidstr] = self
+					agent._objs[context][oidstr] = self
 
 			def value(self):
 				return socket.inet_ntoa(
@@ -392,22 +403,22 @@ class netsnmpAgent(object):
 		# Return an instance of the just-defined class to the agent
 		return IpAddress()
 
-	def Table(self, oidstr, indexes, columns, extendable = False):
+	def Table(self, oidstr, indexes, columns, counterobj = None, extendable = False, context = ""):
 		agent = self
 
 		# Define a Python class to provide access to the table.
 		class Table(object):
-			def __init__(self, oidstr, idxobjs, coldefs, extendable):
+			def __init__(self, oidstr, idxobjs, coldefs, counterobj, extendable, context):
 				# Create a netsnmp_table_data_set structure, representing both
 				# the table definition and the data stored inside it. We use the
 				# oidstr as table name.
-				self._dataset = libnsa.netsnmp_create_table_data_set(
+				self._dataset = libnsX.netsnmp_create_table_data_set(
 					ctypes.c_char_p(oidstr)
 				)
 
 				# Define the table row's indexes
 				for idxobj in idxobjs:
-					libnsa.netsnmp_table_dataset_add_index(
+					libnsX.netsnmp_table_dataset_add_index(
 						self._dataset,
 						idxobj._asntype
 					)
@@ -424,7 +435,7 @@ class netsnmpAgent(object):
 					# trailing zero byte in C strings
 					size = defobj._data_size + 1 if defobj._asntype == ASN_OCTET_STR \
 												 else defobj._data_size
-					result = libnsa.netsnmp_table_set_add_default_row(
+					result = libnsX.netsnmp_table_set_add_default_row(
 						self._dataset,
 						colno,
 						defobj._asntype,
@@ -443,7 +454,8 @@ class netsnmpAgent(object):
 					oidstr,
 					extendable
 				)
-				result = libnsa.netsnmp_register_table_data_set(
+				self._handler_reginfo.contents.contextName = context
+				result = libnsX.netsnmp_register_table_data_set(
 					self._handler_reginfo,
 					self._dataset,
 					None
@@ -456,7 +468,13 @@ class netsnmpAgent(object):
 
 				# Finally, we keep track of all registered SNMP objects for the
 				# getRegistered() method.
-				agent._objs[oidstr] = self
+				agent._objs[context][oidstr] = self
+
+				# If "counterobj" was specified, use it to track the number
+				# of table rows
+				if counterobj:
+					counterobj.update(0)
+				self._counterobj = counterobj
 
 			def addRow(self, idxobjs):
 				dataset = self._dataset
@@ -466,7 +484,7 @@ class netsnmpAgent(object):
 					def __init__(self, idxobjs):
 						# Create the netsnmp_table_set_storage structure for
 						# this row.
-						self._table_row = libnsa.netsnmp_table_data_set_create_row_from_defaults(
+						self._table_row = libnsX.netsnmp_table_data_set_create_row_from_defaults(
 							dataset.contents.default_row
 						)
 
@@ -488,7 +506,7 @@ class netsnmpAgent(object):
 						# do special handling for the trailing zero byte in C strings
 						size = snmpobj._data_size + 1 if snmpobj._asntype == ASN_OCTET_STR \
 													  else snmpobj._data_size
-						result = libnsa.netsnmp_set_row_column(
+						result = libnsX.netsnmp_set_row_column(
 							self._table_row,
 							column,
 							snmpobj._asntype,
@@ -500,11 +518,14 @@ class netsnmpAgent(object):
 
 				row = TableRow(idxobjs)
 
-				libnsa.netsnmp_table_dataset_add_row(
+				libnsX.netsnmp_table_dataset_add_row(
 					dataset,        # *table
 					row._table_row  # row
 				)
 
+				if self._counterobj:
+					self._counterobj.update(self._counterobj.value() + 1)
+
 				return row
 
 			def value(self):
@@ -618,18 +639,33 @@ class netsnmpAgent(object):
 
 				return retdict
 
-		# Return an instance of the just-defined class to the agent
-		return Table(oidstr, indexes, columns, extendable)
+			def clear(self):
+				row = self._dataset.contents.table.contents.first_row
+				while bool(row):
+					nextrow = row.contents.next
+					libnsa.netsnmp_table_dataset_remove_and_delete_row(
+						self._dataset,
+						row
+					)
+					row = nextrow
 
-	def getRegistered(self):
-		""" Returns a dictionary with the currently registered SNMP objects. """
+		# Return an instance of the just-defined class to the agent
+		return Table(oidstr, indexes, columns, counterobj, extendable, context)
+
+	def getContexts(self):
+		return self._objs.keys()
+
+	def getRegistered(self, context = ""):
+		"""Returns a dictionary with the currently registered SNMP objects.
+		Returned is a dictionary objects for the specified `context`, which
+		defaults to the default context ""."""
 		myobjs = {}
-		for (oidstr,snmpobj) in self._objs.iteritems():
+		for oidstr, snmpobj in self._objs[context].iteritems():
 			myobjs[oidstr] = {
 				"type": type(snmpobj).__name__,
 				"value": snmpobj.value()
 			}
-		return myobjs
+		return dict(myobjs)
 
 	def start(self):
 		""" Starts the agent. Among other things, this means connecting
@@ -637,7 +673,7 @@ class netsnmpAgent(object):
 		self._started = True
 		libnsa.init_snmp(self.AgentName)
 
-	def poll(self):
+	def check_and_process(self):
 		""" Blocks and processes incoming SNMP requests. """
 		return libnsa.agent_check_and_process(1)
 
